use crate::config::config::Config;
use crate::store::{
    ignorer::Ignorer,
    indexer::Indexer,
    nsobject::NsObject,
    structs::{self, to_obj_path},
};
use crate::utils::path::to_repo_relative;
use colored::Colorize;
use std::fs;
// use glob::glob;
use std::path::PathBuf;

pub struct AddArgs {
    /// List of files/dirs to add
    pub files: Vec<String>,
    /// Allow ignored files
    pub force: bool,
    /// Select all files
    pub all: bool,
}

pub fn exec(args: AddArgs, config: Config) {
    structs::init(config.get_root_unsafe());

    // Init ignorer
    let mut ignorer = Ignorer::new(config.get_root_unsafe());
    ignorer.active_nsignore(!args.force);

    // Init indexer
    let mut indexer = Indexer::new(config.get_root_unsafe());
    let _ = indexer.load();

    if args.all {
        return add_dir(config.get_root_unsafe(), &mut ignorer, &mut indexer);
    }

    for obj_to_add in args.files.iter() {
        let mut path_to_add = config.execution_path.clone();
        path_to_add.push(PathBuf::from(obj_to_add));

        if path_to_add.exists() {
            // ignore object if it is ns config file or nsignored
            if ignorer.should_ignore(&path_to_add) {
                continue;
            }

            if path_to_add.is_dir() {
                indexer.index_dir(path_to_add.clone());
                add_dir(&path_to_add, &mut ignorer, &mut indexer);
            } else {
                indexer.index_file(path_to_add);
            }
        } else if NsObject::from_local_path(&to_obj_path(&path_to_add)).exists() {
            indexer.index_file(path_to_add);
        } else {
            // try globbing
            // todo!()
            // else panic
            panic!("fatal: pathspec '{}' did not match any files", obj_to_add);
        }
    }

    // print all path ignored
    if ignorer.ignored_paths.len() > 0 {
        println!("The following paths are ignored by one of your .nsignore files:");
        for ignored_path in ignorer.ignored_paths {
            println!("{}", to_repo_relative(&ignored_path).display());
        }
        println!(
            "{}",
            "hint: Use -f if you really want to add them.".bright_red()
        );
        // hint: Turn this message off by running
        // hint: "git config advice.addIgnoredFile false"
    }

    dbg!(indexer.save());
    /*
    for all files
    if globbing
    take all match glob nextsyncignore
    else
    if dir
    add dir
    add all childs with nextsyncignore
    else
    add files
    */
}

/// Add all files in a directory
///
/// # Parameters
/// * `dir`: the directory to add from
fn add_dir(dir: &PathBuf, ignorer: &mut Ignorer, indexer: &mut Indexer) {
    let entries = match fs::read_dir(dir) {
        Ok(entries) => entries,
        Err(err) => {
            eprintln!("Failed to read {} ({err})", dir.display());
            return;
        }
    };

    // Read locals files and folders
    for entry in entries {
        let entry = match entry {
            Ok(entry) => entry,
            Err(err) => {
                eprintln!("Failed to read entry {err}");
                continue;
            }
        };

        // ignore object if it is ns config file or nsignored
        if ignorer.should_ignore(&entry.path()) {
            continue;
        }

        if entry.path().is_dir() {
            indexer.index_dir(entry.path().clone());
            add_dir(&entry.path(), ignorer, indexer);
        } else {
            indexer.index_file(entry.path());
        }
    }
}
