use crate::config::config::Config;
use crate::store::{
    indexer::Indexer,
    nsobject::NsObject,
    object::{Obj, ObjStatus, ObjType},
    structs,
    structs::to_obj_path,
};
use crate::utils::path;
use colored::{ColoredString, Colorize};
use std::collections::HashMap;
use std::fs;
use std::path::PathBuf;
use std::sync::{Arc, Mutex};
use threadpool::ThreadPool;

pub struct StatusArgs {
    pub nostyle: bool,
}

type HashMapObj = HashMap<String, Obj>;
type HashMapObjStatuses = Arc<Mutex<HashMapObj>>;

struct ObjStatuses {
    created: HashMapObjStatuses,
    modified: HashMapObjStatuses,
    deleted: HashMapObjStatuses,
    // staged: Arc<Mutex<Vec<Obj>>>,
    // not_staged: Arc<Mutex<Vec<Obj>>>,
    // untracked: Arc<Mutex<Vec<Obj>>>,
    staged: Vec<Obj>,
    not_staged: Vec<Obj>,
    untracked: Vec<Obj>,
}

impl ObjStatuses {
    fn new() -> Self {
        ObjStatuses {
            created: Arc::new(Mutex::new(HashMap::new())),
            modified: Arc::new(Mutex::new(HashMap::new())),
            deleted: Arc::new(Mutex::new(HashMap::new())),
            // staged: Arc::new(Mutex::new(Vec::new())),
            // not_staged: Arc::new(Mutex::new(Vec::new())),
            // untracked: Arc::new(Mutex::new(Vec::new()))
            staged: Vec::new(),
            not_staged: Vec::new(),
            untracked: Vec::new(),
        }
    }

    fn push_created(&self, key: String, value: Obj) {
        self.created.lock().unwrap().insert(key, value);
    }

    fn push_modified(&self, key: String, value: Obj) {
        self.modified.lock().unwrap().insert(key, value);
    }

    fn push_deleted(&self, key: String, value: Obj) {
        self.deleted.lock().unwrap().insert(key, value);
    }

    fn get_created(&self) -> HashMapObj {
        self.created.lock().unwrap().clone()
    }
}

pub struct ObjStaged {
    pub staged: Vec<Obj>,
    pub not_staged: Vec<Obj>,
}

fn setup_staged(obj_statuses: Arc<ObjStatuses>, indexer: &Indexer) -> ObjStaged {
    let mut staged = Vec::new();
    let mut not_staged = Vec::new();

    for (_, mut obj) in obj_statuses.get_created() {
        obj.set_status(ObjStatus::Created);
        if indexer.is_staged(&obj) {
            staged.push(obj);
        } else {
            not_staged.push(obj);
        }
    }

    ObjStaged { staged, not_staged }

    // for (_, mut obj) in self.modified.lock().unwrap().iter() {
    //     obj.set_status(ObjStatus::Modified);
    //     if indexer.is_staged(&obj) {
    //         self.staged.lock().unwrap().push(obj);
    //     } else {
    //         self.not_staged.lock().unwrap().push(obj);
    //     }
    // }

    // for (_, mut obj)in self.deleted.lock().unwrap().iter() {
    //     obj.set_status(ObjStatus::Deleted);
    //     if indexer.is_staged(&obj) {
    //         self.staged.lock().unwrap().push(obj);
    //     } else {
    //         self.not_staged.lock().unwrap().push(obj);
    //     }
    // }
}

pub fn exec(args: StatusArgs, config: Config) {
    let status = get_obj_changes(&config);
    if args.nostyle {
        todo!("Not style not implemented yet");
    }
    print_status(&status, config.get_root_unsafe());
}

pub fn get_obj_changes(config: &Config) -> ObjStaged {
    structs::init(config.get_root_unsafe());

    // use root of repo if no custom path has been set by the command
    let root = if config.is_custom_execution_path {
        config.execution_path.clone()
    } else {
        config.get_root_unsafe().to_path_buf()
    };

    let indexer = Arc::new({
        let mut indexer = Indexer::new(config.get_root_unsafe());
        indexer.load_unsafe();
        indexer
    });

    let pool = ThreadPool::new(4);
    let repo_root = config.get_root_unsafe().clone();
    let res = Arc::new(ObjStatuses::new());

    let pool_clone = pool.clone();
    let res_clone = Arc::clone(&res);
    let indexer_clone = Arc::clone(&indexer);
    pool.execute(move || {
        compare_dir(pool_clone, indexer_clone, &repo_root, &root, res_clone);
    });

    pool.join();

    for entry in res.created.lock().unwrap().iter() {
        println!("created: {}", entry.0);
    }
    for entry in res.modified.lock().unwrap().iter() {
        println!("modified: {}", entry.0);
    }
    for entry in res.deleted.lock().unwrap().iter() {
        println!("deleted: {}", entry.0);
    }

    // find moved and copied

    // find staged

    setup_staged(Arc::clone(&res), &indexer)
}

///
fn print_status(objs: &ObjStaged, root_path: &PathBuf) {
    if objs.staged.len() == 0 && objs.not_staged.len() == 0 {
        println!("Nothing to push, working tree clean");
        return;
    }

    if objs.staged.len() != 0 {
        println!("Changes to be pushed:");
        println!("  (Use \"nextsync reset\" to unstage)");
        // (use "git restore --staged <file>..." to unstage)
        // by alphabetical order
        for obj in objs.staged.iter() {
            print_object(&obj, root_path, |status: &str| status.green());
        }
    }
    // modified
    // deleted
    // renamed
    // new file
    println!("Changes not staged for push:");
    println!("  (Use \"nextsync add <file>...\" to update what will be pushed)");
    for obj in objs.not_staged.iter() {
        print_object(&obj, root_path, |status: &str| status.red());
    }

    // println!("Untracked files:");
    // for obj in objs.untracked.iter() {
    //     println!("{}", obj.cpy_path());
    // }
}

fn print_object(obj: &Obj, root_path: &PathBuf, color: impl Fn(&str) -> ColoredString) {
    println!(
        "        {}{}",
        match obj.get_status() {
            ObjStatus::Created => match *obj.get_obj_type() {
                ObjType::Blob => color("new file:   "),
                ObjType::Tree => color("new dir:    "),
                ObjType::Obj => color("new:         "),
            },
            ObjStatus::Modified => color("modified:   "),
            ObjStatus::Deleted => color("deleted:    "),
            _ => "unknown".red(),
        },
        color(&path::to_string(&obj.get_env_relative_path(root_path)))
    );
}

///
///
/// * `pool`:
/// * `root_path`: path of the repo's root
/// * `path`: path we should analyze
/// * `res`: the struct in which we should store the response
fn compare_dir(
    pool: ThreadPool,
    indexer: Arc<Indexer>,
    root_path: &PathBuf,
    path: &PathBuf,
    res: Arc<ObjStatuses>,
) {
    let entries = match fs::read_dir(path) {
        Ok(entries) => entries,
        Err(err) => {
            eprintln!("Failed to read {} ({err})", path.display());
            return;
        }
    };

    let mut local_files = HashMap::new();
    let mut local_dirs = HashMap::new();

    // Read locals files and folders
    for entry in entries {
        let entry = match entry {
            Ok(entry) => entry,
            Err(err) => {
                eprintln!("Failed to read entry {err}");
                continue;
            }
        };

        if entry.path().ends_with(".nextsync") {
            continue;
        }

        let local_obj = Obj::from_local_path(&entry.path().into());

        if entry.path().is_dir() {
            if local_obj.is_new() {
                // TODO: opti move files in new directory
                if indexer.is_staged_parent(&local_obj) {
                    let res_clone = Arc::clone(&res);
                    add_childs(root_path, &entry.path(), res_clone);
                } else {
                    res.push_created(local_obj.cpy_path(), local_obj);
                }
            } else {
                // can be modified
                local_dirs.insert(local_obj.cpy_path(), local_obj);

                // Analyze sub folder
                let pool_clone = pool.clone();
                let root_path_clone = root_path.clone();
                let res_clone = Arc::clone(&res);
                let indexer_clone = Arc::clone(&indexer);
                pool.execute(move || {
                    compare_dir(
                        pool_clone,
                        indexer_clone,
                        &root_path_clone,
                        &entry.path(),
                        res_clone,
                    );
                });
            }
        } else {
            if local_obj.is_new() {
                res.push_created(local_obj.cpy_path(), local_obj);
            } else {
                local_files.insert(local_obj.cpy_path(), local_obj);
            }
        }
    }

    // Read ns objects to find deleted
    let entries = NsObject::from_local_path(&to_obj_path(&path));
    for entry in entries.iter() {
        if entry.is_file() {
            match local_files.get(entry.get_obj_path().to_str().unwrap()) {
                None => res.push_deleted(path::to_string(entry.get_obj_path()), {
                    let mut obj = entry.get_local_obj();
                    obj.set_status(ObjStatus::Deleted);
                    obj
                }),
                Some(local_obj) if local_obj != entry => {
                    res.push_modified(local_obj.cpy_path(), local_obj.clone())
                }
                Some(_) => (),
            }
        } else {
            match local_dirs.get(entry.get_obj_path().to_str().unwrap()) {
                None => res.push_deleted(path::to_string(entry.get_obj_path()), {
                    let mut obj = entry.get_local_obj();
                    obj.set_status(ObjStatus::Deleted);
                    obj
                }),
                Some(_) => (),
            }
        }
    }
}

fn add_childs(root_path: &PathBuf, path: &PathBuf, res: Arc<ObjStatuses>) {
    let entries = match fs::read_dir(path) {
        Ok(entries) => entries,
        Err(err) => {
            eprintln!("Failed to read {} ({err})", path.display());
            return;
        }
    };
    for entry in entries {
        let entry = match entry {
            Ok(entry) => entry,
            Err(err) => {
                eprintln!("Failed to read entry {err}");
                continue;
            }
        };

        let local_obj = Obj::from_local_path(&entry.path().into());
        if entry.path().is_dir() {
            if entry.path().ends_with(".nextsync") {
                continue;
            }

            let res_clone = Arc::clone(&res);
            add_childs(root_path, &entry.path(), res_clone);
            res.push_created(local_obj.cpy_path(), local_obj);
        } else {
            res.push_created(local_obj.cpy_path(), local_obj);
        }
    }
}
