use crate::services::service::Service;
use crate::store::object::Obj;
use crate::config::config::Config;
use serde::Deserialize;
use serde_xml_rs::from_str;

pub enum Props {
    CreationDate,
    LastModified,
    ETag,
    ContentType,
    RessourceType,
    ContentLength,
    ContentLanguage,
    DisplayName,
    FileId,
    Permissions,
    Size,
    HasPreview,
    Favorite,
    CommentsUnread,
    OwnerDisplayName,
    ShareTypes,
    ContainedFolderCount,
    ContainedFileCountm,
}


#[derive(Deserialize, Debug)]
struct Propstat {
    #[serde(rename = "prop")]
    prop: Prop,
}

#[derive(Deserialize, Debug)]
struct Prop {
    #[serde(rename = "getlastmodified")]
    last_modified: Option<String>,
    #[serde(rename = "getcontentlength")]
    content_length: Option<u64>,
}

#[derive(Deserialize, Debug)]
struct Response {
    #[serde(rename = "href")]
    href: String,
    #[serde(rename = "propstat")]
    propstat: Propstat,
}

#[derive(Deserialize, Debug)]
struct Multistatus {
    #[serde(rename = "response")]
    responses: Vec<Response>,
}


impl From<&Props> for &str {
    fn from(variant: &Props) -> Self {
        match variant {
            Props::CreationDate => "<d:creationdate />",
            Props::LastModified => "<d:getlastmodified />",
            _ => todo!("Props conversion not implemented"),
        }
    }
}

pub struct ReqProps {
    service: Service,
    properties: Vec<Props>,
}

impl ReqProps {
    pub fn new(path: &str) -> Self {
        let mut service = Service::new();
        service.propfind(path.to_owned());
        ReqProps {
            service,
            properties: Vec::new(),
        }
    }

    pub fn set_config(mut self, config: &Config) -> Self {
        self.service.set_config(config.clone());
        self
    }

    pub fn get_properties(mut self, properties: Vec<Props>) -> Self {
        self.properties.extend(properties);
        self
    }

    pub fn get_property(&mut self, property: Props) {
        self.properties.push(property);
    }

    fn get_body(&self) -> String {
        let mut xml = String::from(
            r#"<?xml version="1.0" encoding="UTF-8"?><d:propfind xmlns:d="DAV:" xmlns:oc="http://owncloud.org/ns" xmlns:nc="http://nextcloud.org/ns"><d:prop>"#,
        );
        for property in &self.properties {
            xml.push_str(property.into());
        }
        xml.push_str(r#"</d:prop></d:propfind>"#);
        xml
    }

    pub async fn send(&mut self) {
        dbg!("send");
        let res = self.service.send().await;
        dbg!("Sent");
        let text = res.unwrap().text().await.unwrap();
        self.parse(&text);
    }

    fn parse (&self, xml: &str) {
        dbg!("Parsing");
        let multistatus: Multistatus = from_str(&xml).unwrap();
        dbg!(multistatus);
    }
}

impl From<Obj> for ReqProps {
    fn from(obj: Obj) -> Self {
        ReqProps::new(obj.get_obj_path().to_str().unwrap())
    }
}
